#!/usr/bin/env python
#
# This python script does the following for each argument:
#    --big     extracts 1d spectra from 2d spectra
#    --smf     finds new positions of the objects creates an SMF file
#
# Requires big type file (equispec) for --big 
# and an extensions type of file for --smf
#
# Last modified 7-28-06

from basics import *
from FITS import *
from Multipack import *
from VTKHelperFunctions import *
import pgplot as PG
import glob,sys,time
import cPickle
from scipy.optimize import leastsq
from scipy.interpolate import splrep,splint

def help():
    print
    print "This python script extracts 1d spectra"
    print "It takes the 2d spectra from Gus's routine extract-2dspec,"
    print "which is extensions format, and outputs to equispec format."
    print
    print "-a  --aperture     APNUM   size of aperture in pixels"
    print "    --big                  1d extraction mode"
    print "-e  --edge         ENUM    edge to exclude"
    print "-f  --flat                 extracts flat"
    print "-h  --help                 print program options"
    print "-k  --knum         KNUM    iteration, used for pipeline"
    print "-l  --lamp                 extracts lamp"
    print "-m  --mask         MSTR    mask file (smf) input"
    print "-ms --mark-spectra         mark position of object spectra"
    print "-nf --no-find              skips position adjustment for extraction"
    print "-o  --orders       ORDNUM  orders of fit"
    print "-p  --pth          PNUM    s/n threshold for trace (default 2)"
    print "-s  --science              extracted science (2dspec)"
    print "-se --shift-edge           shift slit edges with objects"
    print "-t  --trace-all            traces all objects"
    print "-to --trace-order  TONUM   order of the global curvature trace"
    print "    --smf                  smf mode"
    print "-v  --verb                 verbose ouput"
    print 

def param(f,x,h,verb=0):
    x0,x1 = clip([h-3,h+4],0,x[-1])
    wt = greater(f,0)
    wt2 = asarray([wt,wt])
    kz = nonzero(wt)
    if kz:
        wstatus = 1
        j = kz[argmin(abs(kz-h))]
        wt = 1.0*VTKSeedConnectivity(wt2,x=[j],y=[0])[0]
        #wt = VTKDilate(wt,5,numret=1)
        #qdump("wt.fits",1.0*wt)
        #qdump("swt.fits",1.0*swt)
        if verb: print f
        if verb: print h
        if verb: print wt
        h = divz(sum(x*wt),sum(wt))
        h = int(h+0.5)
        if verb: print h
        x0,x1 = clip([h-3,h+4],0,x[-1])

    ### wt fails kz is a zero array!!!
    else: wstatus = 0

    if x1-x0 < 7:
       if x0 == 0: x1 = 7
       if x1 == x[-1]: x0 = x1-7
    j = argmax(f[x0:x1]) + x0
    c = x[j]
    a = f[j]
    if a > 0: a = log10(a)
    else: a = 2.0
    s = log10(1.5)
    inside = between(x,x0,x1)
    outside = logical_not(inside)
    b = median(compress(outside,f))
    if verb: print c,a,s,b
    return [c,a,s,b],wstatus

def gauss(param,f,x,verb=0,fitting=1):
    c,a,s,b = param
    a = pow(10,a)
    s = pow(10,s)
    if verb: print c,a,s,b
    m = a*getexp(-0.5*((x-c)/s)**2) + b
    if fitting: return f - m
    else: return m

def fit(f,x,h):
    fit = leastsq(gauss,param,args=(x,f))
    return fit

def GetTrace((Spec2D,Noise2D),params,instrument,edge,torder=2):
    y, x = indices(Spec2D.shape,Float32)
    #if "LDSS3" in instrument: ReallyBad = less(y,5)
    #else: ReallyBad = logical_or(less(y,2),greater(y,y.shape[0]-3))
    EdgeMask = logical_or(less(y,edge),greater(y,y.shape[0]-(1+edge)))
    y = y - params[0]
    x0, x1 = 0.0, 1.0*x.shape[1]
    xc, xw = (x1+x0)/2.0, (x1-x0)/2.0
#    w = divz(Spec2D,Noise2D)
    w = getsqrt(Spec2D)*greater(Noise2D,0)
    #w = w*greater(w,0.0)*logical_not(ReallyBad)
    w = w*greater(w,0.0)*logical_not(EdgeMask)
    w = w*between(y-sum(y.flat*w.flat)/sum(w.flat),-10,+10)
    b = legendre((x.flat-xc)/xw,torder)
    t = svdfit(transpose(b)*w.flat[::,NewAxis],y.flat*w.flat)
    #sys.exit()
    return t
#    Test the solution by rebinning spectra:
#    p = reshape(dot(t,b),x.shape)
#    r = VTKImageTransform(Spec2D,0.0*p,p,reverse=0,numret=1)


def GetProfile(D,edge,verb=0,mark=0):
    (Spec2D,Noise2D) = D["data"]
    y, x = indices(Spec2D.shape,Float32)
    EdgeMask = logical_or(less(y,edge),greater(y,y.shape[0]-(1+edge)))
    w = Spec2D*logical_not(EdgeMask)
    #qdump("w.fits",w)
    h = D["pos"]
    ob = D["id"]
    p = median(transpose(w))
    #p = median(transpose(Spec2D))
    x = arange(p.shape[0])
    guess,wstatus = param(p,x,h,verb)
    lsq = leastsq(gauss,guess,(p,x,verb),full_output=1,xtol=1e-4,ftol=1e-4,gtol=1e-4,epsfcn=1e-4,factor=0.1)
    status = lsq[2]
    params = lsq[0]
    if verb>0:
        PG.newbox(extrema(x),[ min(p)-(max(p)-min(p))/10, max(p)+(max(p)-min(p))/3],ch=2.3)
        PG.xlabel("Y",ch=2.3)
        PG.ylabel("Profile",ch=2.3)
        PG.drawln(x,p,ls=1,hist=1)
        xa = arange(min(x),max(x)+1,0.01,Float)
        PG.drawln(x,p,ls=1,ci=2,hist=1)
        PG.drawln(xa,gauss(params,p,xa,0,0),ls=1,ci=4,hist=0)
        PG.drawln([params[0],params[0]],[min(p)-2*(max(p)-min(p)),max(p)+2*(max(p)-min(p))],ls=1,ci=7)

    elif mark>0:
        W = divz(1,Noise2D).flat**2
        F = Spec2D.flat
        cf = divz(sum(F*W),sum(W))
        cf = 0.0
        rf = sqrt(divz(sum(W*(F-cf)**2),sum(W)))
        a1,a2 = cf-3*rf, +3*rf
        So = sort(F)
        No = len(So)
        a1,a2 = So[int(0.05*No+1)], So[int(0.95*No+1)]
        c1,c2 = 0,Spec2D.shape[1]-1
        xfm = array([-0.5,1.0,0.0,-0.5,0.0,1.0])
        kcur = None
        xcur = Spec2D.shape[1]/2
        ycur = params[0]
        while kcur not in ["a","M","q"]:
            PG.cpgsvp(0.05,0.20,0.15,0.90)
            PG.newbox([0,Spec2D.shape[0]-1],[min(p)-(max(p)-min(p))/10,max(p)+(max(p)-min(p))/3],ch=2.3,manvp=1)
            PG.xlabel("Y",ch=2.3)
            PG.ylabel("Profile",ch=2.3)
            PG.drawln(x,p,ls=1,hist=1)
            xa = arange(min(x),max(x)+1,0.01,Float)
            PG.drawln(x,p,ls=1,ci=2,hist=1)
            PG.drawln(xa,gauss(params,p,xa,0,0),ls=1,ci=4,hist=0)
            PG.drawln([params[0],params[0]],[min(p)-2*(max(p)-min(p)),max(p)+2*(max(p)-min(p))],ls=1,ci=7)
            PG.subtitle(D["id"],ch=2.3)
            PG.cpgsvp(0.25,0.95,0.15,0.90)
            PG.newbox([c1,c2],[0,Spec2D.shape[0]-1],ch=2.3,manvp=1,era=0)
            PG.xlabel("X",ch=2.3)
            PG.ylabel("Y",ch=2.3)
            PG.subtitle(D["id"],ch=2.3)
            PG.cpggray(Spec2D,Spec2D.shape[1],Spec2D.shape[0],\
                    1,Spec2D.shape[1],1,Spec2D.shape[0],a1,a2,xfm)
            PG.drawln([0,Spec2D.shape[1]],[0.5+params[0],0.5+params[0]],ls=1,ci=7)
            PG.drawln([0,Spec2D.shape[1]],[0.5+params[0]-10**params[2],0.5+params[0]-10**params[2]],ls=2,ci=7)
            PG.drawln([0,Spec2D.shape[1]],[0.5+params[0]+10**params[2],0.5+params[0]+10**params[2]],ls=2,ci=7)
            PG.drawln([0,Spec2D.shape[1]],[0.5+h,0.5+h],ls=1,ci=2)
            PG.drawln([0,Spec2D.shape[1]],[0.5+h+mark/2.,0.5+h+mark/2.],ls=2,ci=2)
            PG.drawln([0,Spec2D.shape[1]],[0.5+h-mark/2.,0.5+h-mark/2.],ls=2,ci=2)
            kcur,(xcur,ycur),(xdum,ydum) = PG.getcurs(xcur,ycur,xfm)
            if kcur == "I": raise "UserInterrupt"
            if kcur in ["a","q"]: break
            if kcur.upper() == "M":
                params = array([ydum,99,3,0.0])
                status = 1
                if kcur == "M": break
          
            aran = a2-a1
            xran = c2-c1
            if kcur == "j": params[0] = max([0.0, params[0] - 0.5])
            if kcur == "k": params[0] = min([params[0] + 0.5, Spec2D.shape[0]-1])
            if kcur == "h":
                c1 = c1 - xran/10.
                if c1 < 0: c1 = 0
                c2 = c1 + xran
            if kcur == "l":
                c2 = c2 + xran/10.
                if c2 > Spec2D.shape[1]-1: c2 = Spec2D.shape[1]-1
                c1 = c2 - xran
          
            zf = 2.0
            if kcur == "s": a1=a1+aran/4.; a2=a2-aran/4.
            if kcur == "d": a1=a1-aran/4.; a2=a2+aran/4.
            if kcur == "f": a1=a1-aran/4.; a2=a2-aran/4.
            if kcur == "b": a1=a1+aran/4.; a2=a2+aran/4.
            if kcur == "r": a1,a2 = So[int(0.05*No+1)], So[int(0.95*No+1)]; c1,c2 = 0,Spec2D.shape[1]-1
            if kcur == "x" or kcur == "X":
                if kcur == "x": xran = xran*zf; c1 = xcur - xran/2.; c2 = xcur + xran/2.
                if kcur == "X": xran = xran/zf; c1 = xcur - xran/2.; c2 = xcur + xran/2.
                if c1 < 0: c1 = 0
                if c2 > Spec2D.shape[1]-1: c2 = Spec2D.shape[1]-1
            PG.cpgeras()
      
        ### cheat we subtract a -1 to offset the -1 we would get in stage-extract-1dspec   
        params[0] -= 1.0



    if not (0 < params[0] < x[-1]): status = 0
    params[1] = pow(10,params[1])
    params[2] = pow(10,params[2])
    fmt = "%12s %8.1f %4d" + 4 * "%9.1f" + " %r %r"
    out = (ob,D["flux"],h,) + tuple(params) + (status,wstatus,)
    print fmt % out
    D["info"] = params
    D["status"] = status
    if verb>1: raw_input("pause! ")
    return params,status

def getdata(h,refobjs,testsmf,instrument=None):
    f = FITS(fitsfile)
    f.goto(h)
    D = {"id":f["OBJECT"].strip(),"pos":f["CNTRLINE"]}
    if D["id"] in refobjs: D["ref"] = True
    else: D["ref"] = False
    if not testsmf or D["ref"]:
       D["data"] = f.data()
       D["data"] = where(between(D["data"],-1e38,1e38),D["data"],0.0)
    if D["ref"]:
       median_profile = median(transpose(D["data"][0]))
       median_profile = median_profile - getpct(median_profile,10)/2.0
       D["flux"] = sum(median_profile)
       print "%15s %10.2f" % (D["id"], D["flux"])
    f.close()
    return D 

def getbigdata(s,refobjs,testsmf):
    f = FITS(fitsfile)
    ff = FITS(fitsfile.replace("_big.fits","_sig.fits"))
    D = {"id":f["OBJ%03d" % (s)].strip(), \
       "pre":f["CNTRL%03d" % (s)], \
       "pos":f["CNTRL%03d" % (s)]-f["CSECT%dA" % (s)]+1}
    if D["id"] in refobjs: D["ref"] = True
    else: D["ref"] = False
    if not testsmf or D["ref"]:
        ll = [0,f["CSECT%dA" % (s)]-1]
        ur = [f["NAXIS1"]-1,f["CSECT%dB" % (s)]-1]
        D["data"] = f.readsec(ll,ur),ff.readsec(ll,ur)
        D["data"] = where(between(D["data"],-1e38,1e38),D["data"],0.0)
    if D["ref"]:
       median_profile = median(transpose(D["data"][0]))
       median_profile = median_profile - getpct(median_profile,10)/2.0
       D["flux"] = sum(median_profile)
       print "%15s %10.2f" % (D["id"], D["flux"])
    f.close()
    ff.close()
    return D 
 
def getextractdata(h,refobjs,file):
    f = FITS(file)
    f.goto(h)
    D = {"id":f["OBJECT"].strip(),"pos":f["CNTRLINE"],"data":f.data()}
    D["data"] = where(between(D["data"],-1e38,1e38),D["data"],0.0)
    f.close()
    if D["id"] in refobjs: D["ref"] = True
    else: D["ref"] = False
    return D 

def Etrace(o,d,p,t,a,r=2,shift=True,f=False):
    if not f: sys.stdout.write(" %r" % (o)); sys.stdout.flush()
    if f: s = d
    else: (s,n) = d
    x = arange(s.shape[1])
    y2,x2 = indices(s.shape,Float32)
    y2 = y2 - t
    if shift:
       sm = VTKMedian(s,7,1,numret=1)
       for j in range(3):
         u = between(y2,p-max([8,a/2.0]),p+max([8,a/2.0]))
         u = u * between(x2, s.shape[1]/4., 3*s.shape[1]/4.)
         sp = add.reduce(u*sm)
         np = sqrt(add.reduce(u*power(n,2)))
         sn = divz(sp,np)
         sn = compress(sn,sn)
         if len(sn) > len(s)/3: sn = getpct(sn,75)
         else: sn = 0.0
         if sn > r:
           um = u*not_equal(s,0)*not_equal(n,0)
           um = um*sm
#           um = um*divz(sm,n)
           p2 = divz(sum((sm*um*y2).flat), sum((sm*um).flat))
           p2 = clip(p2,p-max([8,a]),p+max([8,a]))
           sys.stdout.write("(%.0f,%1.f,%.1f)" % (sn,p,p2-p)); sys.stdout.flush()
           p = p2
       else:
         sys.stdout.write("(%.0f,%1.f,XXX)" % (sn,p)); sys.stdout.flush()
    spec = zeros(x.shape,Float64)
    noise = zeros(x.shape,Float64)
    y2 = y2 - p
    hi = greater(y2,-a/2.0)
    hi2 = greater(y2,-a/2.0-1) - hi
    hip = -(y2+a/2.0) * hi2
    lo = less(y2,a/2.0)
    lo2 = less(y2,a/2.0+1) - lo
    lop = (y2-a/2.0) * lo2
    mid = lo*hi
    use = (mid+hip+lop)
    intspec = lambda d: add.reduce(d*use)
    if f: return (intspec(s),p)
    else: return (intspec(s),getsqrt(intspec(power(n,2))),p)

def E(s,p,a):
    p = int(p+0.5)
    dims = [s.shape,s.shape[1:]][len(s.shape)==3]
    spec = zeros((dims[1],),Float64)
    noise = zeros((dims[1],),Float64)
    if 0 < p < dims[0]:
       bot, top = p-a/2, p+a/2
       bot,top = clip([bot,top],0,dims[0])
       if len(s.shape) == 3:
          spec = sum(s[0][bot:top])
          noise = sqrt(sum(power(s[1][bot:top],2)))
       else:
          spec = sum(s[bot:top])
    return (spec,noise)

def spec1d(fitsfile,smffile,a,pth,flat,ofile,testsmf=False,testbig=False,orders=None,\
        knum=None,edge=0,trace_all=0,verb=0,mark=0,no_find=0,torder=1,lamp=None,shifte=0):

    L_all = open(smffile,"r").readlines()
    L = [l.split() for l in L_all]

    instrument = [l[1] for l in L if l[0] == "INSTRUMENT"][0]

    refobjs = [l[1] for l in L if l[0] == "HOLE"]
    smf_ids = [l[1] for l in L if l[0] in ["HOLE","SLIT"]]
    smf_x = array([float(l[4]) for l in L if l[0] in ["HOLE","SLIT"]])
    smf_y = array([float(l[5]) for l in L if l[0] in ["HOLE","SLIT"]])

    t0 = time.time()
    f = FITS(fitsfile)
    if testbig:
        D = [getbigdata(s,smf_ids,testsmf) for s in 1+arange(f["N_SLITS"])]
    else:
        D = [getdata(s,smf_ids,testsmf,instrument) for s in 1+arange(f["N_SLITS"])]

    fluxes = array([d["flux"] for d in D if d["ref"]])
    flux_cut = 0.0
    f.close()
    t1 = time.time()
    print "data read in %.3fs" % (t1-t0)
    print

    t0 = time.time()

    if mark:
       print "(a)ccept the position"
       print "(m)ark the position"
       print "adjust the marked position up (k) or down (j)\n"
       print "(s)harper"
       print "(d)uller"
       print "(b)righter"
       print "(f)ainter\n"
       print "zoom out (x)"
       print "zoom in (X)"
       print "move plot to the left (h) or right (l)\n"
       PG.opendev("/xs",nx=1,ny=1,land=0,size=10,asp=0.2,ask=0)
       [GetProfile(d,edge,verb=verb,mark=a) for d in D]
       PG.closedev()
    else:
       if verb: PG.opendev("/xs",nx=1,ny=1,land=1,size=6.5,ask=0)
       [GetProfile(d,edge,verb=verb,mark=0) for d in D if d["ref"] and d["flux"] > flux_cut]
       if verb: PG.closedev()

    for d in D: d["ref"] = (d["flux"] > flux_cut)

    t1 = time.time()
    print "gaussians fit to half the objects in %.3fs" % (t1-t0)
    print

    sigmas = array([d["info"][2] for d in D if d["ref"]])
    status = array([d["status"] for d in D if d["ref"]])
    dpos = array([d["info"][0]-d["pos"] for d in D if d["ref"]])
    print "median(dpos)=",median(dpos)
    amps = array([d["info"][1] for d in D if d["ref"]])
    pflux = array([d["flux"] for d in D if d["ref"]])
    refx = array([smf_x[smf_ids.index(d["id"])] for d in D if d["ref"]])
    refy = array([smf_y[smf_ids.index(d["id"])] for d in D if d["ref"]])
    refo = array([(d["id"] in refobjs) for d in D if d["ref"]])

    good = greater(amps,0.0)*equal(status,1.0)

    refo,refx,refy,pflux,amps,dpos,sigmas,status=compress(good,[refo,refx,refy,pflux,amps,dpos,sigmas,status],1)

    ampcut = 0.0
    weights = greater(amps,ampcut)
    weights = weights*between(sigmas,0.1,2*median(sigmas))

    def getmar(x,mar=0.5):
       x1, x2 = getpct(x,10), getpct(x,90)
       xs = (x2-x1)
       return x1-mar*xs,x2+mar*xs

    ### Plot to file

    if orders: ox,oy = orders
    else: ox,oy = 0,0

    if not mark and (testsmf and not orders): user_input = "go"
    else: user_input = "done"

    pgdev = "/xs"

    ### Loop through user inputed orders
    # Fit line to flux and amp plot...
    while pgdev[:2] == "/x":

       if user_input == "done":
         if testbig: pgdev = "%s/vcps" %(fitsfile.replace("_big.fits","_1spec_leastsq.ps"))
         else: pgdev = "%s/vcps" %(fitsfile.replace("_2spec.fits","_1spec_leastsq.ps"))
       PG.opendev(pgdev,nx=3,ny=2,land=0,asp=2.0/3.0,size=9)

       pps = where(refo,4,17)

       PG.newbox(getmar(pflux),divide(getmar(pflux),2*sqrt(2*pi*median(sigmas)**2)),ch=2)
#       PG.newbox((getmar(pflux)[0],getmar(pflux,3)),(getmar(amps)[0],getmar(amps,3)),ch=2)
       PG.xlabel("F",ch=2)
       PG.ylabel("A",ch=2)
       PG.drawpts(pflux,amps,pps)

       PG.newbox(add([-30,30],extrema(refx)),[0,4],ch=2)
       PG.xlabel("X",ch=2)
       PG.ylabel("\gs",ch=2)
       PG.drawpts(refx,sigmas,pps)

       PG.newbox(add([-30,30],extrema(refx)),add([-30,30],extrema(refy)),ch=2)
       PG.xlabel("X",ch=2)
       PG.ylabel("Y",ch=2)
       MMperPixel = 0.2 * 0.34499
       [PG.arrow([refx[jo],refx[jo]], [refy[jo],refy[jo]+dpos[jo]*4],ch=0.4,ls=1,lw=1,ci=1) for jo in range(len(refx))]

       PG.newbox(add([-30,30],extrema(refx)),[-20,20],ch=2)
       PG.xlabel("X",ch=2)
       PG.ylabel("\gDY",ch=2)
       PG.drawpts(refx,dpos,pps,ch=2)

       lrefx = refx/300.0
       lrefy = refy/300.0
       lxb = legendre(lrefx,ox)
       lyb = legendre(lrefy,oy)
       basis_fns = [xb*yb for xb in lxb for yb in lyb]

       uweights = 1*weights

       nfit = 1
       for fitpass in range(nfit):
          sol = linear_least_squares(transpose(basis_fns)*uweights[::,NewAxis],dpos*uweights)[0]
          if not fitpass:
             sol = planefit(compress(uweights,transpose(basis_fns),0),compress(uweights,dpos),sol,minfunc=mad)[0]
          if type(sol) is FloatType: sol = array([sol])
          fitted = dot(sol,basis_fns)
          resid = dpos - fitted
          if fitpass < nfit-1:
            uweights = uweights*less(abs(resid)/bwt(compress(uweights,resid))[1],5.0)
          scatter = sqrt(sum(power(resid,2)*uweights)/sum(uweights))
          print "SOL=",sol
          print "RMS=",scatter
          print "BWT=",bwt(compress(uweights,resid))

       PG.drawpts(refx,resid,pps,ch=2,ci=2)
       PG.drawpts(compress(uweights,refx),compress(uweights,resid),compress(uweights,pps),ch=2,ci=4)

       scatter = sqrt(sum(power(resid,2)*uweights)/sum(uweights))
       print "SOL=",sol
       print "RMS=",scatter
       print "BWT=",bwt(compress(uweights,resid))
       slabel0 = "(ox,oy,cut)= (%d,%d,%f)" % (ox,oy,ampcut)
       PG.wrttext("",slabel0,-1.5,1.6,0.05,"T",1,0)
       slabel1 = "SOL = %s" % (sol)
       PG.wrttext("",slabel1,-2.5,1.6,0.05,"T",1,0)
       slabel2 = "RMS = %.6f" % (scatter)
       PG.wrttext("",slabel2,-3.5,1.6,0.05,"T",1,0)

       PG.newbox(add([-30,30],extrema(refy)),[-20,20],ch=2)
       PG.xlabel("Y",ch=2)
       PG.ylabel("\gDY",ch=2)
       PG.drawpts(refy,resid,pps)

       PG.newbox(add([-30,30],extrema(refx)),add([-30,30],extrema(refy)),ch=2)
       PG.xlabel("X",ch=2)
       PG.ylabel("Y",ch=2)
       MMperPixel = 0.2 * 0.34499
       [PG.arrow([refx[jo],refx[jo]], [refy[jo],refy[jo]+resid[jo]*4],ch=0.4,ls=1,lw=1,ci=1) for jo in range(len(refx))]

       PG.closedev()

       if pgdev[:2] == "/x":
          prompt = "<(ox,oy,cut) = (%d,%d,%f)> " % (ox,oy,ampcut)
          user_input = raw_input( prompt)
          try:
             ox,oy,ampcut = eval(user_input)
             weights = greater(amps,ampcut)
             weights = weights*between(sigmas,0.1,2*median(sigmas))
          except: pass

    if testsmf:
        if not os.path.exists(ofile):
            o = open(ofile,"w")
            o.write("# stage2\n")
            o.write("ox%i = %i\n" % (knum,ox))
            o.write("oy%i = %i\n" % (knum,oy))
            o.close()
        else:
            olist = open(ofile,"r").readlines()
            odict = {}
            for ol in olist:
                if ol[0] != "#":
                    on = ol.split("=")
                    odict[on[0].strip()]=on[1].strip()
            odict["ox%i" % (knum)] = "%i" % (ox)
            odict["oy%i" % (knum)] = "%i" % (oy)
            olist = [on for on in odict]
            olist.sort()
            o = open(ofile,"w")
            o.write("# stage2\n")
            [o.write("%s = %s\n" % (on,odict[on])) for on in olist]
            o.close()

    # Correcting postions based fit
    lrefx = smf_x/300.0
    lrefy = smf_y/300.0
    lxb = legendre(lrefx,ox)
    lyb = legendre(lrefy,oy)
    basis_fns = [xb*yb for xb in lxb for yb in lyb]

    dpos = dot(sol,basis_fns)
    for d in D:
        j = smf_ids.index(d["id"])
        d["orig_pos"] = 1*d["pos"]
        if not mark and not no_find:
            d["pos"] = d["pos"] + dpos[j]

    if not testsmf:
        # Trace atmospheric dispersion (only applies to FITS)
        
        for d in D:
           if d["ref"] and (trace_all or d["id"] in refobjs) and (d["info"][1] > 0):
              d["info"][0] = d["pos"]
              d["trace"] = GetTrace(d["data"],d["info"],instrument,edge,torder=torder)
           else:
              d["trace"] = None

        traces = asarray([d["trace"] for d in D if d["trace"]])
        trace = mean(traces)
        trace[0] = 0.0

        print
        for j in range(len(traces)):
           print "Trace #%d = %r" % (j,traces[j])
        print "Trace=",trace
        print

    msigma = add.reduce(uweights*sigmas)/add.reduce(uweights)
    #print sigmas
    print "msigma=",msigma
    print


    ### SMF mode
    if testsmf:
        smfhead = [l.rstrip()+'\n' for l in L_all if l.split()[0] not in ["HOLE","SLIT"]]
        smflines = [l for l in L_all if l.split()[0] in ["HOLE","SLIT"]]
       
        hole_or_slit = lambda w: w=="HOLE"

        smf_i = [l[1] for l in L if l[0] in ["HOLE","SLIT"]]
        smf_x = array([float(l[4]) for l in L if l[0] in ["HOLE","SLIT"]])
        smf_y = array([float(l[5]) for l in L if l[0] in ["HOLE","SLIT"]])
        smf_l = array([float(l[[7,8][hole_or_slit(l[0])]]) for l in L if l[0] in ["HOLE","SLIT"]])
        smf_r = array([float(l[[8,9][hole_or_slit(l[0])]]) for l in L if l[0] in ["HOLE","SLIT"]])

        if mark:
           pixel_shifts = []
           for i in smf_i:
              d = D[smf_ids.index(i)]
              pixel_shift = d["info"][0] - d["pos"]
              pixel_shifts += [ pixel_shift]
           pixel_shifts = array( pixel_shifts)
        else:
           lrefx = smf_x/300.0
           lrefy = smf_y/300.0
           lxb = legendre(lrefx,ox)
           lyb = legendre(lrefy,oy)
           basis_fns = [xb*yb for xb in lxb for yb in lyb]
           pixel_shifts = dot(sol,basis_fns)

        if "IMACS" in instrument:
          if "short" in instrument:
#       rescale pixel shifts into mm in the focal plane
#       for f/2 the scale is (1/2.89856) mm per arcsec
            MMperPixel = 0.2 * 0.34500
            PixelperMM = 1/(0.2 * 0.34500)
          elif "long" in instrument:
            MMperPixel = 0.111 * 0.34500
            PixelperMM = 1/(0.111 * 0.34500)
        elif "LDSS3" in instrument:
            MMperPixel = 0.189 * (1/2.88)
            PixelperMM = 1/(0.189 * (1/2.88))
        mm_shifts = pixel_shifts * MMperPixel
        
        #print "exclude (mm): ",5*MMperPixel
        
        nsmffile = fitsfile.replace("_2spec.fits",".SMF")
        fsmf = open(nsmffile,"w")
        [fsmf.write(l) for l in smfhead]
        
        for q,l in enumerate(smflines):
            n_line = l.split()
            n_x = "%.3f" % (smf_x[q] + (mm_shifts[q]))
            n_l = "%.3f" % (smf_l[q] + (mm_shifts[q]))
            n_r = "%.3f" % (smf_r[q] - (mm_shifts[q]))


            # How do we deal with sumspec and uneven spectra being extracted???
            #if float(n_l)*PixelperMM < 5.0 or float(n_r)*PixelperMM < 5.0:
            if float(n_l) < 0.0 or float(n_r) < 0.0:
                print "%15s pos failed %10s, %10s" % (n_line[1],n_l,n_r)
                continue

            scol = lambda q,r: " ".join([n_line[z] for z in range(q,r+1)])
            if shifte:
                if n_line[0] in ["HOLE"]: n3_line=" ".join([scol(0,3),n_x,scol(5,9)," ",n_line[10]])
                elif n_line[0] in ["SLIT"]: n3_line=" ".join([scol(0,3),n_x,scol(5,8)," ",n_line[9]])
            else:    
                if n_line[0] in ["HOLE"]: n3_line=" ".join([scol(0,3),n_x,scol(5,7),n_l,n_r," ",n_line[10]])
                elif n_line[0] in ["SLIT"]: n3_line=" ".join([scol(0,3),n_x,scol(5,6),n_l,n_r," ",n_line[9]])

#            print n3_line
            fsmf.write(n3_line+"\n")
        fsmf.close()
        print
        print nsmffile

    ### Big mode
    elif not testsmf:
        print "Reading %s" % (flat)
        f = FITS(flat)
        FD = [getextractdata(s,refobjs,flat) for s in 1+arange(f["N_SLITS"])]
        f.close()
        print "Done reading %s" % (flat)

        if lamp:
            print "Reading %s" % (lamp)
            f = FITS(lamp)
            LD = [getextractdata(s,refobjs,lamp) for s in 1+arange(f["N_SLITS"])]
            f.close()
            print "Done reading %s" % (lamp)

        for fd in FD:
            if fd["ref"]: fd["ap"] = a
            else: fd["ap"] = a
            j = smf_ids.index(fd["id"])
            if not no_find: fd["pos"] = fd["pos"] + dpos[j]

        if lamp:
            for ld in LD:
                if ld["ref"]: ld["ap"] = a
                else: ld["ap"] = a
                j = smf_ids.index(ld["id"])
                if not no_find: ld["pos"] = ld["pos"] + dpos[j]

        # Spectral Extractions
        for d in D:
          if d["ref"]: d["ap"] = a
          else: d["ap"] = a
       
        naxis1 = D[0]["data"][0].shape[1]
        x0, x1 = 0.0, 1.0*naxis1
        xc, xw = (x1+x0)/2.0, (x1-x0)/2.0
        b = legendre((arange(0,naxis1,1,Float32)-xc)/xw,len(trace)-1)
        tfun = dot(trace,b)
        t = open(fitsfile.replace("_big.fits",".trace"),"wb")
        cPickle.dump(tfun,t) 
        t.close()

        #print D

        t0 = time.time()
 
        print "Starting extractions"
        S = [Etrace(d["id"],d["data"],d["pos"],tfun,d["ap"],r=pth,shift=not no_find,f=False) for d in D]
        print
        Sp = array([s[2] for s in S])
        S = array([[s[0] for s in S],[s[1] for s in S]])

        # Flattened-Flatfield Extractions

        Sids = [d["id"] for d in D]
        Fids = [fd["id"] for fd in FD]
        uFD = [FD[Fids.index(Sid)] for Sid in Sids]

        print "Starting flat extractions"
        FS = [Etrace(fd["id"],fd["data"],fd["pos"],tfun,fd["ap"],r=0.0,shift=False,f=True) for fd in uFD]
        FSp = array([s[1] for s in FS])
        FS = array([s[0] for s in FS])
        c = concatenate([S,[FS]],0)

        # Lamp Extractions
        if lamp:
            Lids = [ld["id"] for ld in LD]
            uLD = [LD[Lids.index(Sid)] for Sid in Sids]

            print "Starting lamp extractions"
            LS = [Etrace(ld["id"],ld["data"],ld["pos"],tfun,ld["ap"],r=0.0,shift=False,f=True) for ld in uLD]
            LSp = array([s[1] for s in LS])
            LS = array([s[0] for s in LS])
            c = concatenate([c,[LS]],0)
        
        #########################
        #########################
        #########################
        t1 = time.time()
        print "Done with all extractions in %.3fs" % (t1-t0)
        print
        
        naxes = array(c.shape)[::-1].tolist()

        if testbig:
            f = FITS(fitsfile,readonly=0)
            apnums = [f["APNUM%d" % (1+j)].split() for j in range(len(D))]
            apcens = array([f["cntrl%03d" % (1+j)] + (1 + Sp[j] - D[j]["pos"]) for j in range(len(D))])
#            apcens = array([f["csect%da" % (1+j)] + Sp[j] for j in range(len(D))])
            aplo,aphi = apcens[NewAxis,::] + 0.5*a*array([-1,1])[::,NewAxis]
            for j in range(len(D)): apnums[j][2:4] = "%.2f" % (aplo[j]), "%.2f" % (aphi[j])
            fixpos = lambda j: f["cntrl%03d" % (1+j)] + (1 + Sp[j] - D[j]["orig_pos"])
#            fixpos = lambda j: f["csect%da" % (j+1)] + Sp[j]
            [f.modkey("excen%03d" % (1+j), fixpos(j)) for j in range(len(D))]
            f.close()

        if testbig: ffitsfile = fitsfile.replace("_big.fits","_2p_1spec.fits")
        else: ffitsfile = fitsfile.replace("_2spec.fits","_1spec.fits")

        ff=FITS(ffitsfile,readonly=0,create=1)
        f = FITS(fitsfile)
        if not testbig: f.goto(1)
        ff.copyheader(f)
        f.close()
        ff.close()

        ff=FITS(ffitsfile,readonly=0)
        ff.resize(naxes)
        [ff.modkey(key,val) for (key,val) in [("dapert",a)]]
        if testbig: [ff.modkey("APNUM%d" % (1+j)," ".join(apnums[j])) for j in range(len(D))]

        ff.flush()
        ff.putdata(c)
        ff.close()
        print ffitsfile

ofile = "orders.dat"
fitsfile = None
flat = None
smffile = None
testsmf = False
testbig = False
orders = None
knum = None
lamp = None
a = None
pth = 1
n = 0
verb = 0
mark = 0
no_find = 0
edge = 0
torder = 2
trace_all = 0
shifte = 0

for arg in sys.argv:
    if arg in ("-a","--aperture"): a = int(sys.argv[n+1])
    elif arg in ("-m","--mask"): smffile = sys.argv[n+1]
    elif arg in ("-l","--lamp"): lamp = sys.argv[n+1]
    elif arg in ("-k","--knum"): knum = int(sys.argv[n+1])
    elif arg in ("-o","--orders"): orders = eval(sys.argv[n+1])
    elif arg in ("-to","--trace-order"): torder = eval(sys.argv[n+1])
    elif arg in ("-e","--edge"): edge = int(sys.argv[n+1])
    elif arg in ("-f","--flat"): flat = sys.argv[n+1]
    elif arg in ("-p","--pth"): pth = float(sys.argv[n+1])
    elif arg in ("-s","--science"): fitsfile = sys.argv[n+1]
    elif arg in ("--smf"): testsmf = True
    elif arg in ("--big"): testbig = True
    elif arg in ("-t","--trace-all"): trace_all = 1
    elif arg in ("-v","--verb"): verb += 1
    elif arg in ("-ms","--mark-spectra"): mark += 1
    elif arg in ("-nf","--no-find"): no_find += 1
    elif arg in ("-se --shift-edge"): shifte = 1
    elif arg in ("-h","--help"):
        help()
        sys.exit()
    n += 1

if a is not None and fitsfile is not None and smffile is not None:
    spec1d(fitsfile,smffile,a,pth,flat,ofile,testsmf,testbig=testbig,orders=orders,\
           knum=knum,edge=edge,trace_all=trace_all,verb=verb,mark=mark,no_find=no_find,\
           torder=torder,lamp=lamp,shifte=shifte)
else:
    help()
    sys.exit()


